import requests
import webbrowser
import time
import sys
import base64
import logging

from bs4 import BeautifulSoup

LOG = logging.getLogger(__name__)


def device_auth(org_domain, oidc_client_id):
    LOG.debug('Started device auth')
    url = "https://" + org_domain + "/oauth2/v1/device/authorize"
    payload = 'client_id=' + oidc_client_id + \
              '&scope=openid%20okta.apps.sso%20okta.apps.read'
    headers = {'Content-Type': 'application/x-www-form-urlencoded'}

    request = requests.post(url, headers=headers, data=payload)
    response = request.json()

    verification_url = response['verification_uri_complete']
    webbrowser.open_new_tab(verification_url)

    LOG.debug(f'Got device code {response["device_code"]}')
    return response['device_code']


def verification_and_token(org_domain, oidc_client_id, device_code):
    LOG.debug('Started verification of device code')
    url = "https://" + org_domain + "/oauth2/v1/token"
    payload = 'client_id=' + oidc_client_id + '&device_code=' + device_code + \
              '&grant_type=urn%3Aietf%3Aparams%3Aoauth%3Agrant-type%3Adevice_code'
    headers = {'Content-Type': 'application/x-www-form-urlencoded'}

    while True:
        request = requests.post(url, headers=headers, data=payload)
        response = request.json()

        # Check for authorization pending
        if request.status_code == 400 and response['error'] == 'authorization_pending':
            LOG.debug('Waiting for verification')
            time.sleep(5)
            continue

        # Check for successful verification
        if request.status_code == 200:
            break

        # Unexpected state. Die.
        LOG.error(response)
        sys.exit(1)

    LOG.debug('Validated device code and got access_token & id_token')
    return response['access_token'], response['id_token']


def session_and_token(org_domain, oidc_client_id, access_token, id_token, aws_acct_fed_app_id):
    LOG.debug('Started getting of session token')
    url = "https://" + org_domain + "/oauth2/v1/token"
    payload = 'client_id=' + oidc_client_id + '&actor_token=' + access_token + \
              '&actor_token_type=urn%3Aietf%3Aparams%3Aoauth%3Atoken-type%3Aaccess_token' + \
              '&subject_token=' + id_token + \
              '&subject_token_type=urn%3Aietf%3Aparams%3Aoauth%3Atoken-type%3Aid_token' + \
              '&grant_type=urn%3Aietf%3Aparams%3Aoauth%3Agrant-type%3Atoken-exchange' + \
              '&requested_token_type=urn%3Aokta%3Aoauth%3Atoken-type%3Aweb_sso_token' + \
              '&audience=urn%3Aokta%3Aapps%3A' + aws_acct_fed_app_id
    headers = {'Content-Type': 'application/x-www-form-urlencoded'}

    request = requests.post(url, headers=headers, data=payload)
    response = request.json()

    LOG.debug('Got session token')
    return response['access_token']


def saml_assertion(org_domain, session_token):
    LOG.debug('Started SAML assertion')
    # Get SAML assertion
    url = 'https://' + org_domain + '/login/token/sso?token=' + session_token
    response = requests.get(url)

    # Extract response value from SAML assertion call
    parser = BeautifulSoup(response.text, "html.parser")
    saml_response = parser.find("input", {"name": "SAMLResponse"})['value']
    parser = BeautifulSoup(base64.b64decode(saml_response), features="xml")

    # Retrieve list of Roles from the SAML assertion
    roles = {}
    role_xml = parser.find("saml2:Attribute", {"Name": "https://aws.amazon.com/SAML/Attributes/Role"})
    for role in role_xml.find_all("saml2:AttributeValue"):
        idp_and_role = role.text.split(',')
        roles[idp_and_role[1]] = idp_and_role[0]

    session_duration_xml = parser.find("saml2:Attribute",
                                       {"Name": "https://aws.amazon.com/SAML/Attributes/SessionDuration"})
    session_duration = session_duration_xml.find("saml2:AttributeValue").text

    LOG.debug(f'Got valid SAML response: {saml_response}')
    return saml_response, roles, int(session_duration)



import boto3
import logging

from botocore.utils import JSONFileCache

LOG = logging.getLogger(__name__)

class TokenGenerationException(Exception):
    """Exception for credential not having token"""
    pass


def credentials(profile_name, region_name):
    LOG.debug(f'Creating boto3 session for profile {profile_name} and region {region_name}')
    session = boto3.Session(profile_name=profile_name, region_name=region_name)

    __update_credential_provider_cache(session)

    creds = session.get_credentials()
    if creds.token is None:
        LOG.error("No session credential found")
        raise TokenGenerationException()

    return creds, session.region_name


def __update_credential_provider_cache(session):
    """Setting up a custom cache implementation like aws cli"""

    cred_chain = session._session.get_component('credential_provider')
    json_file_cache = JSONFileCache()

    def _update(provider_name):
        cred_chain.get_provider(provider_name).cache = json_file_cache

    provider_for_cache = [
        'assume-role',
        'assume-role-with-web-identity',
        'sso'
    ]

    [_update(each_provider) for each_provider in provider_for_cache]



import argparse
import json
import keyring
import sys


def setup(subparsers, parent_parser):
    common_parser = argparse.ArgumentParser(add_help=False)
    common_parser.add_argument('--access-key', required=True, help='AWS access key')
    common_parser.add_argument('--account-id', required=False, help='AWS Account ID for the name')
    common_parser.add_argument('--username', required=False, help='Username of a AWS user associated with the access key for the name')

    store_summary = 'Store IAM user access key and secret key securely for streamlined authentication'
    store_parser = subparsers.add_parser('store-iam-user-credentials', description=store_summary, help=store_summary, parents=[parent_parser, common_parser])
    store_parser.set_defaults(func=run_store)
    store_parser.add_argument('--secret-key', required=True, help='AWS secret key')
    
    get_summary = 'Retrieve IAM user credentials for AWS CLI profiles or application authentication'
    get_parser = subparsers.add_parser('get-iam-user-credentials', description=get_summary, help=get_summary, parents=[parent_parser, common_parser])
    get_parser.set_defaults(func=run_get)
    get_parser.add_argument('--credential-process', action='store_true', help='Output the credential in AWS credential process syntax')

    
def run_store(args):
    service_name = '-'.join(filter(None, ['aws', args.account_id, args.username]))
    keyring.set_password(service_name, args.access_key, args.secret_key)


def run_get(args):
    service_name = '-'.join(filter(None, ['aws', args.account_id, args.username]))
    secret_key = keyring.get_password(service_name, args.access_key)

    if args.credential_process:
        # https://docs.aws.amazon.com/cli/latest/userguide/cli-configure-sourcing-external.html
        print(json.dumps({
            "Version": 1,
            "AccessKeyId": args.access_key,
            "SecretAccessKey": secret_key
        }))
    else:
        command = '$env:' if sys.platform == 'win32' else 'export '

        print(f'{command}AWS_ACCESS_KEY_ID="{args.access_key}"')
        print(f'{command}AWS_SECRET_ACCESS_KEY="{secret_key}"')
    
    


import os
import pyperclip
import webbrowser
import sys
import logging

from ..aws.session import credentials
from ..aws.api import signin_url

LOG = logging.getLogger(__name__)


def setup(subparsers, parent_parser):
    summary = 'Open a web browser for graphical access to the AWS Console'
    parser = subparsers.add_parser('open-browser', description=summary, help=summary, parents=[parent_parser])
    parser.set_defaults(func=run)

    parser.add_argument('-P', '--profile', default=os.getenv("AWS_PROFILE"), help="The AWS profile to create the pre-signed URL with")
    parser.add_argument('-R', '--region', default=os.getenv("AWS_REGION", os.getenv("AWS_DEFAULT_REGION")), help="The AWS Region to send the request to")

    no_browser_group = parser.add_mutually_exclusive_group()
    no_browser_group.add_argument('--clip', action='store_true', help="don't open the web browser, but copy the signin URL to clipboard")
    no_browser_group.add_argument('--stdout', action='store_true', help="don't open the web browser, but echo the signin URL to stdout")


def run(args):
    creds, region_name = credentials(args.profile, args.region)
    url = signin_url(creds, region_name)
    LOG.debug(f'Generated aws console URL: {url}')

    if args.clip:
        pyperclip.copy(url)
    elif args.stdout:
        print(url)
    elif not webbrowser.open_new_tab(url):
        LOG.error('No browser found on the system')
        sys.exit("No browser found. Try --help for other options")



import os
import configparser
import sys
import logging

LOG = logging.getLogger(__name__)


def setup(subparsers, parent_parser):
    summary = 'Initilize fusion app with creation of aws fusion alias'
    parser = subparsers.add_parser('init', description=summary, help=summary, parents=[parent_parser])
    parser.set_defaults(func=run)

def run(args):
    cli_dir = os.path.expanduser(os.path.join('~', '.aws', 'cli'))

    def create_alias(config: configparser.ConfigParser):
        LOG.debug('Creating fusion aws alias')
        if not config.has_section('toplevel'):
            config.add_section('toplevel')
        config['toplevel']['fusion'] = f'!{sys.executable} -m aws_fusion.app'

    def update_aws_cli_alias_file():
        if not os.path.isdir(cli_dir):
            LOG.debug(f"Path {cli_dir} doesn't exists, creating" )
            os.makedirs(cli_dir)
        cli_alias_full_path = os.path.join(cli_dir, 'alias')
        config = configparser.ConfigParser()

        if os.path.isfile(cli_alias_full_path):
            LOG.debug(f'Found alias file {cli_alias_full_path}')
            config.read(cli_alias_full_path)

        create_alias(config)

        with os.fdopen(os.open(cli_alias_full_path, os.O_WRONLY | os.O_CREAT, 0o600), 'w') as f:
            f.truncate()
            config.write(f)
            LOG.debug(f'Updated alias file {cli_alias_full_path}')

    update_aws_cli_alias_file()    
